#include <cmath>
#include <cstdio>

#include <math/MathUtils.h>

#include "vecmath/Quat4f.h"
#include "vecmath/Vector3f.h"
#include "vecmath/Vector4f.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

Quat4f::Quat4f()
{
	m_afElements[0] = 0;
	m_afElements[1] = 0;
	m_afElements[2] = 0;
	m_afElements[3] = 0;
}

Quat4f::Quat4f( float w, float x, float y, float z )
{
	m_afElements[0] = w;
	m_afElements[1] = x;
	m_afElements[2] = y;
	m_afElements[3] = z;
}

Quat4f::Quat4f( const Quat4f& rq )
{
	m_afElements[0] = rq.m_afElements[0];
	m_afElements[1] = rq.m_afElements[1];
	m_afElements[2] = rq.m_afElements[2];
	m_afElements[3] = rq.m_afElements[3];
}

Quat4f& Quat4f::operator = ( const Quat4f& rq )
{
	if( this != ( &rq ) )
	{
		m_afElements[0] = rq.m_afElements[0];
		m_afElements[1] = rq.m_afElements[1];
		m_afElements[2] = rq.m_afElements[2];
		m_afElements[3] = rq.m_afElements[3];
	}
    return( *this );
}

Quat4f::Quat4f( const Vector3f& v )
{
	m_afElements[0] = 0;
	m_afElements[1] = v[0];
	m_afElements[2] = v[1];
	m_afElements[3] = v[2];
}

Quat4f::Quat4f( const Vector4f& v )
{
	m_afElements[0] = v[0];
	m_afElements[1] = v[1];
	m_afElements[2] = v[2];
	m_afElements[3] = v[3];
}

const float& Quat4f::operator [] ( int i ) const
{
	return m_afElements[ i % 4 ];
}

float& Quat4f::operator [] ( int i )
{
	return m_afElements[ i % 4 ];
}

float Quat4f::w() const
{
	return m_afElements[0];
}

float Quat4f::x() const
{
	return m_afElements[1];
}

float Quat4f::y() const
{
	return m_afElements[2];
}

float Quat4f::z() const
{
	return m_afElements[3];
}

Vector3f Quat4f::xyz() const
{
	return Vector3f
		(
			m_afElements[0],
			m_afElements[1],
			m_afElements[2]
		);
}

float Quat4f::abs() const
{
	return sqrt
		(
			m_afElements[0] * m_afElements[0] +
			m_afElements[1] * m_afElements[1] +
			m_afElements[2] * m_afElements[2] +
			m_afElements[3] * m_afElements[3]
		);
			
}

float Quat4f::absSquared() const
{
	return
		(
			m_afElements[0] * m_afElements[0] +
			m_afElements[1] * m_afElements[1] +
			m_afElements[2] * m_afElements[2] +
			m_afElements[3] * m_afElements[3]
		);
}

void Quat4f::normalize()
{
	float absVal = abs();

	m_afElements[0] /= absVal;
	m_afElements[1] /= absVal;
	m_afElements[2] /= absVal;
	m_afElements[3] /= absVal;
}

Quat4f Quat4f::normalized() const
{
	float absVal = abs();

	return Quat4f
		(
			m_afElements[0] / absVal,
			m_afElements[1] / absVal,
			m_afElements[2] / absVal,
			m_afElements[3] / absVal
		);
}

void Quat4f::conjugate()
{
	m_afElements[1] = -m_afElements[1];
	m_afElements[2] = -m_afElements[2];
	m_afElements[3] = -m_afElements[3];
}

Quat4f Quat4f::conjugated() const
{
	return Quat4f
		(
			m_afElements[0],
			-m_afElements[1],
			-m_afElements[2],
			-m_afElements[3]
		);
}

void Quat4f::invert()
{
	Quat4f inverse = conjugated() * ( 1.0f / absSquared() );

	m_afElements[ 0 ] = inverse.m_afElements[ 0 ];
	m_afElements[ 1 ] = inverse.m_afElements[ 1 ];
	m_afElements[ 2 ] = inverse.m_afElements[ 2 ];
	m_afElements[ 3 ] = inverse.m_afElements[ 3 ];
}

Quat4f Quat4f::inverse() const
{
	return conjugated() * ( 1.0f / absSquared() );
}

void Quat4f::getAxisAngle( float* pfRadians, float* pfx, float* pfy, float* pfz )
{
	float theta = acos( w() ) * 2;
	float vectorNorm = sqrt( x() * x() + y() * y() + z() * z() );

	*pfRadians = theta;
	*pfx = x() / vectorNorm;
	*pfy = y() / vectorNorm;
	*pfz = z() / vectorNorm;
}

void Quat4f::getAxisAngle( float* pfRadians, Vector3f& rvAxis )
{
	float theta = acos( w() ) * 2;
	float vectorNorm = sqrt( x() * x() + y() * y() + z() * z() );

	*pfRadians = theta;
	rvAxis[0] = x() / vectorNorm;
	rvAxis[1] = y() / vectorNorm;
	rvAxis[2] = z() / vectorNorm;	
}

void Quat4f::setAxisAngle( float fRadians, float fx, float fy, float fz )
{
	m_afElements[0] = cos( fRadians / 2 );

	float sinHalfTheta = sin( fRadians / 2 );
	float vectorNorm = sqrt( fx * fx + fy * fy + fz * fz );

	m_afElements[1] = fx * sinHalfTheta / vectorNorm;
	m_afElements[2] = fy * sinHalfTheta / vectorNorm;
	m_afElements[3] = fz * sinHalfTheta / vectorNorm;
}

void Quat4f::setAxisAngle( float fRadians, const Vector3f& crAxis )
{
	m_afElements[0] = cos( fRadians / 2 );

	float sinHalfTheta = sin( fRadians / 2 );
	float vectorNorm = crAxis.abs();

	m_afElements[1] = crAxis.x() * sinHalfTheta / vectorNorm;
	m_afElements[2] = crAxis.y() * sinHalfTheta / vectorNorm;
	m_afElements[3] = crAxis.z() * sinHalfTheta / vectorNorm;
}

void Quat4f::print()
{
	printf( "< %1.2f + %1.2f i + %1.2f j + %1.2f k >\n",
		m_afElements[0], m_afElements[1], m_afElements[2], m_afElements[3] );
}

// static
float Quat4f::dot( const Quat4f& q0, const Quat4f& q1 )
{
	return
		(
			q0.w() * q1.w() +
			q0.x() * q1.x() +
			q0.y() * q1.y() +
			q0.z() * q1.z()
		);
}

// static
Quat4f Quat4f::lerp( const Quat4f& q0, const Quat4f& q1, float alpha )
{
	return( ( q0 + alpha * ( q1 - q0 ) ).normalized() );
}

// static
Quat4f Quat4f::slerp( const Quat4f& q0, const Quat4f& q1, float alpha, float cosOmegaThreshold )
{
	float cosOmega = Quat4f::dot( q0, q1 );

	// if they're too close, then just lerp them
	if( cosOmega > cosOmegaThreshold )
	{
		return Quat4f::lerp( q0, q1, alpha );
	}
	else
	{
		cosOmega = MathUtils::clampToRangeFloat( cosOmega, -1, 1 );
		if( cosOmega < -1 )
		{
			cosOmega = -1;
		}
		if( cosOmega > 1 )
		{
			cosOmega = 1;
		}

		float omega0 = acos( cosOmega ); // original angle between q0 and q1
		float omega = omega0 * alpha; // new angle

		Quat4f q2 = q1 - q0 * cosOmega;
		q2.normalize();

		return( q0 * cos( omega ) + q2 * sin( omega ) );
	}
}

// static
Quat4f Quat4f::randomRotation( float u0, float u1, float u2 )
{
	float z = u0;
	float theta = 2.f * MathUtils::PI * u1;
	float r = sqrt( 1.f - z * z );
	float w = MathUtils::PI * u2;

	return Quat4f
	(
		cos( w ),
		sin( w ) * cos( theta ) * r,
		sin( w ) * sin( theta ) * r,
		sin( w ) * z
	);
}

Vector3f Quat4f::rotateVector( const Vector3f& v )
{
	// return q * v * q^-1
	return ( ( *this ) * Quat4f( v ) * conjugated() ).xyz();
}

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

Quat4f operator + ( const Quat4f& q0, const Quat4f& q1 )
{
	return Quat4f
		(
			q0.w() + q1.w(),
			q0.x() + q1.x(),
			q0.y() + q1.y(),
			q0.z() + q1.z()
		);
}

Quat4f operator - ( const Quat4f& q0, const Quat4f& q1 )
{
	return Quat4f
		(
			q0.w() - q1.w(),
			q0.x() - q1.x(),
			q0.y() - q1.y(),
			q0.z() - q1.z()
		);
}

Quat4f operator * ( const Quat4f& q0, const Quat4f& q1 )
{
	return Quat4f
		(
			q0.w() * q1.w() - q0.x() * q1.x() - q0.y() * q1.y() - q0.z() * q1.z(),
			q0.w() * q1.x() + q0.x() * q1.w() + q0.y() * q1.z() - q0.z() * q1.y(),
			q0.w() * q1.y() - q0.x() * q1.z() + q0.y() * q1.w() + q0.z() * q1.x(),
			q0.w() * q1.z() + q0.x() * q1.y() - q0.y() * q1.x() + q0.z() * q1.w()
		);
}

Quat4f operator * ( float f, const Quat4f& q )
{
	return Quat4f
		(
			f * q.w(),
			f * q.x(),
			f * q.y(),
			f * q.z()
		);
}

Quat4f operator * ( const Quat4f& q, float f )
{
	return Quat4f
		(
			f * q.w(),
			f * q.x(),
			f * q.y(),
			f * q.z()
		);
}
