#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstdlib>

#include "vecmath/Vector2f.h"
#include "vecmath/Vector2d.h"
#include "vecmath/Vector2i.h"
#include "vecmath/Vector3f.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

Vector2f::Vector2f()
{
	m_afElements[0] = 0;
	m_afElements[1] = 0;
}

Vector2f::Vector2f( float x, float y )
{
	m_afElements[0] = x;
	m_afElements[1] = y;
}

Vector2f::Vector2f( const Vector2f& rv )
{
	m_afElements[0] = rv.m_afElements[0];
	m_afElements[1] = rv.m_afElements[1];
}

Vector2f::Vector2f( const Vector2d& rv )
{
	m_afElements[0] = static_cast< float >( rv.x() );
	m_afElements[1] = static_cast< float >( rv.y() );
}

Vector2f::Vector2f( const Vector2i& rv )
{
	m_afElements[0] = rv.x();
	m_afElements[1] = rv.y();
}

Vector2f& Vector2f::operator = ( const Vector2f& rv )
{
	if( this != &rv )
	{
		m_afElements[0] = rv.m_afElements[0];
		m_afElements[1] = rv.m_afElements[1];
	}
	return *this;
}

Vector2f& Vector2f::operator = ( const Vector2d& rv )
{
	m_afElements[0] = static_cast< float >( rv.x() );
	m_afElements[1] = static_cast< float >( rv.y() );

	return *this;
}

Vector2f& Vector2f::operator = ( const Vector2i& rv )
{
	m_afElements[0] = rv.x();
	m_afElements[1] = rv.y();
	
	return *this;
}

const float& Vector2f::operator [] ( int i ) const
{
	return m_afElements[ i % 2 ];
}

float& Vector2f::operator [] ( int i )
{
	return m_afElements[ i % 2 ];
}

float Vector2f::x() const
{
	return m_afElements[0];
}

float Vector2f::y() const
{
	return m_afElements[1];
}

Vector2f Vector2f::xy() const
{
	return Vector2f( m_afElements[0], m_afElements[1] );
}

Vector2f Vector2f::yx() const
{
	return Vector2f( m_afElements[1], m_afElements[0] );
}

Vector2f Vector2f::xx() const
{
	return Vector2f( m_afElements[0], m_afElements[0] );
}

Vector2f Vector2f::yy() const
{
	return Vector2f( m_afElements[1], m_afElements[1] );
}

float Vector2f::abs() const
{
	return sqrt( m_afElements[0] * m_afElements[0] + m_afElements[1] * m_afElements[1] );
}

float Vector2f::absSquared() const
{
	return( m_afElements[0] * m_afElements[0] + m_afElements[1] * m_afElements[1] );
}

void Vector2f::normalize()
{
	float norm = sqrt( m_afElements[0] * m_afElements[0] + m_afElements[1] * m_afElements[1] );
	m_afElements[0] = m_afElements[0] / norm;
	m_afElements[1] = m_afElements[1] / norm;
}

Vector2f Vector2f::normalized() const
{
	float length = abs();
	return Vector2f( m_afElements[0] / length, m_afElements[1] / length );
}

void Vector2f::negate()
{
	m_afElements[0] = -m_afElements[0];
	m_afElements[1] = -m_afElements[1];
}

// ---- Utility ----
Vector2f::operator const float* ()
{
	return m_afElements;
}

void Vector2f::print() const
{
	printf( "< %1.2f, %1.2f >\n",
		m_afElements[0], m_afElements[1] );
}

// static
float Vector2f::dot( const Vector2f& v0, const Vector2f& v1 )
{
	return v0.x() * v1.x() + v0.y() * v1.y();
}

//static
Vector3f Vector2f::cross( const Vector2f& v0, const Vector2f& v1 )
{
	return Vector3f
		(
			0,
			0,
			v0.x() * v1.y() - v0.y() * v1.x()
		);
}

// static
Vector2f Vector2f::lerp( const Vector2f& v0, const Vector2f& v1, float alpha )
{
	return alpha * ( v1 - v0 ) + v0;
}

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

Vector2f operator + ( const Vector2f& v0, const Vector2f& v1 )
{
	return Vector2f( v0.x() + v1.x(), v0.y() + v1.y() );
}

Vector2f operator - ( const Vector2f& v0, const Vector2f& v1 )
{
	return Vector2f( v0.x() - v1.x(), v0.y() - v1.y() );
}

Vector2f operator * ( const Vector2f& v0, const Vector2f& v1 )
{
	return Vector2f( v0.x() * v1.x(), v0.y() * v1.y() );
}

Vector2f operator / ( const Vector2f& v0, const Vector2f& v1 )
{
	return Vector2f( v0.x() / v1.x(), v0.y() / v1.y() );
}

Vector2f operator - ( const Vector2f& v )
{
	return Vector2f( -v.x(), -v.y() );
}

Vector2f operator * ( float f, const Vector2f& v )
{
	return Vector2f( f * v.x(), f * v.y() );
}

Vector2f operator * ( const Vector2f& v, float f )
{
	return Vector2f( f * v.x(), f * v.y() );
}
